import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.init import xavier_uniform_

import math

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# Implementation is based on https://github.com/Chen-Yang-Liu/RSICC
# Publication: Remote Sensing Image Change Captioning With Dual-Branch Transformers: A New Method and a Large Scale Dataset
# DOI: 10.1109/TGRS.2022.3218921
# Minor changes have been made to adapt it to our codebase


class CrossTransformer(nn.Module):
    """
    Cross Transformer layer
    """

    def __init__(self, dropout, d_model=512, n_head=12):
        """
        :param dropout: dropout rate
        :param d_model: dimension of hidden state
        :param n_head: number of heads in multi head attention
        """
        super(CrossTransformer, self).__init__()
        self.attention = nn.MultiheadAttention(d_model, n_head, dropout=dropout)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)

        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)
        self.activation = nn.ReLU()

        self.linear1 = nn.Linear(d_model, d_model * 4)
        self.linear2 = nn.Linear(d_model * 4, d_model)

    def forward(self, input1, input2):
        # dif_as_kv
        dif = input2 - input1
        output_1, attn_weight_1 = self.cross(input1, dif)  # (Q,K,V)
        output_2, attn_weight_2 = self.cross(input2, dif)  # (Q,K,V)

        return (output_1, attn_weight_1), (output_2, attn_weight_2)

    def cross(self, input, dif):
        # RSICCformer_D (diff_as_kv)
        attn_output, attn_weight = self.attention(input, dif, dif)  # (Q,K,V)

        output = input + self.dropout1(attn_output)

        output = self.norm1(output)
        ff_output = self.linear2(self.dropout2(self.activation(self.linear1(output))))
        output = output + self.dropout3(ff_output)
        output = self.norm2(output)
        return output, attn_weight
    
    # def forward(self, input1, input2):
    #     dif = input2 - input1

    #     output_1, attn_weight_1 = self.cross(input2, input1, dif)  # (Q, K, V)
    #     output_2, attn_weight_2 = self.cross(input1, input2, dif)  # (Q, K, V)

    #     return (output_1, attn_weight_1), (output_2, attn_weight_2)

    # def cross(self, query, key, value):
    #     attn_output, attn_weight = self.attention(query, key, value)

    #     output = query + self.dropout1(attn_output)
    #     output = self.norm1(output)

    #     ff_output = self.linear2(self.dropout2(self.activation(self.linear1(output))))
    #     output = output + self.dropout3(ff_output)
    #     output = self.norm2(output)

    #     return output,attn_weight


class resblock(nn.Module):
    """
    module: Residual Block
    """

    def __init__(self, inchannel, outchannel, stride=1, shortcut=None):
        super(resblock, self).__init__()
        self.left = nn.Sequential(
            nn.Conv2d(inchannel, int(outchannel / 2), kernel_size=1),
            # nn.LayerNorm(int(outchannel/2),dim=1),
            nn.BatchNorm2d(int(outchannel / 2)),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                int(outchannel / 2),
                int(outchannel / 2),
                kernel_size=3,
                stride=1,
                padding=1,
            ),
            # nn.LayerNorm(int(outchannel/2),dim=1),
            nn.BatchNorm2d(int(outchannel / 2)),
            nn.ReLU(inplace=True),
            nn.Conv2d(int(outchannel / 2), outchannel, kernel_size=1),
            # nn.LayerNorm(int(outchannel / 1),dim=1)
            nn.BatchNorm2d(outchannel),
        )
        self.right = shortcut

    def forward(self, x):
        out = self.left(x)
        residual = x
        out += residual
        return F.relu(out)


class MCCFormers_diff_as_Q(nn.Module):
    """
    RSICCFormers_diff
    """

    def __init__(self, feature_dim, dropout, h, w, d_model=512, n_head=12, n_layers=3):
        """
        :param feature_dim: dimension of input features
        :param dropout: dropout rate
        :param d_model: dimension of hidden state
        :param n_head: number of heads in multi head attention
        :param n_layer: number of layers of transformer layer
        """
        super(MCCFormers_diff_as_Q, self).__init__()
        self.d_model = d_model

        self.n_layers = n_layers

        self.embedding_1D = nn.Embedding(h * w, int(d_model))

        self.projection = nn.Linear(feature_dim, d_model)

        self.transformer = nn.ModuleList(
            [CrossTransformer(dropout, d_model, n_head) for i in range(n_layers)]
        )

        self.resblock = nn.ModuleList(
            [resblock(d_model * 2, d_model * 2) for i in range(n_layers)]
        )

        self.LN = nn.ModuleList([nn.LayerNorm(d_model * 2) for i in range(n_layers)])

        self._reset_parameters()

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        for p in self.parameters():
            if p.dim() > 1:
                xavier_uniform_(p)

    def forward(self, img_feat1, img_feat2):
        batch = img_feat1.size(0)
        num_patches = img_feat1.size(1)

        if img_feat1.size(2) != self.d_model:
            img_feat1 = self.projection(img_feat1)
            img_feat2 = self.projection(img_feat2)

        pos_embedding = (
            self.embedding_1D(torch.arange(num_patches, device=device).to(device))
            .unsqueeze(0)
            .repeat(batch, 1, 1)
        )

        img_feat1 = img_feat1 + pos_embedding
        img_feat2 = img_feat2 + pos_embedding

        encoder_output1 = img_feat1.permute(
            1, 0, 2
        )  # (num_patches, batch_size, d_model)
        encoder_output2 = img_feat2.permute(
            1, 0, 2
        )  # (num_patches, batch_size, d_model)

        output1 = encoder_output1
        output2 = encoder_output2
        output1_list = list()
        output2_list = list()

        for l in self.transformer:
            output1, output2 = l(
                output1[0] if type(output1) == tuple else output1,
                output2[0] if type(output2) == tuple else output2,
            )

            output1_list.append(output1[0])
            output2_list.append(output2[0])

        # MBF
        i = 0
        output = torch.zeros((196, batch, self.d_model * 2)).to(device)
        # change her to ViTb/16 to ViT b/32
        # 196 -> 49
        for res in self.resblock:
            inp = torch.cat([output1_list[i], output2_list[i]], dim=-1)
            output = output + inp
            output = output.permute(1, 2, 0).view(batch, self.d_model * 2, 14, 14) #14,14-> 7,7
            output = res(output)
            output = output.view(batch, self.d_model * 2, -1).permute(2, 0, 1)
            output = self.LN[i](output)
            i = i + 1

        return output


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pe", pe)

        self.embedding_1D = nn.Embedding(52, int(d_model))

    def forward(self, x):
        # fixed
        x = x + self.pe[: x.size(0), :]
        # learnable
        # x = x + self.embedding_1D(torch.arange(52, device=device).to(device)).unsqueeze(1).repeat(1,x.size(1),  1)
        return self.dropout(x)
