import torch
from args import config
from torchvision import transforms
from torch.utils.data import Dataset
import json
import os
import numpy as np
from PIL import Image
import random
import rasterio

dubai_cc_train_transform = {
    "RGB": transforms.Compose(
        [
            transforms.Normalize(mean=[0.618, 0.456, 0.364], std=[0.299, 0.203, 0.172]),
        ]
    ),
    "PCA": transforms.Compose(
        [
            transforms.Normalize(mean=[0.253, 0.632, 0.651], std=[0.057, 0.119, 0.185]),
        ]
    ),
    "MULTISPECTRAL": transforms.Compose([
        # transforms.RandomResizedCrop(size=(50,50), scale=(0.8,1)),
        transforms.Normalize (mean = [0.207,0.297,0.409,0.391,0.432,0.360], std=[0.129,0.180,0.259,0.248,0.284,0.233])

    ]),
}

dubai_cc_test_transform = {
    "RGB": transforms.Compose(
        [
            transforms.Normalize(mean=[0.618, 0.456, 0.364], std=[0.299, 0.203, 0.172]),
        ]
    ),
    "PCA": transforms.Compose(
        [
            transforms.Normalize(mean=[0.253, 0.632, 0.651], std=[0.057, 0.119, 0.185]),
        ]
    ),
    "MULTISPECTRAL": transforms.Compose(
        [

        transforms.Normalize (mean = [0.207,0.297,0.409,0.391,0.432,0.360], std=[0.129,0.180,0.259,0.248,0.284,0.233])
        ]
    )
}


no_change_sentences = [
    "Nothing has changed",
    "No change occurred",
    "Nothing has changed in this area",
    "There is no change to mention",
    "There is no difference",
    "Everything remains the same",
    "The area appears the same",
    "No change in this area",
    "All remained the same",
    "No difference in this area",
]


class DubaiDataset(Dataset):
    def __init__(self, split: str, version="RGB", transform=None, seed=42):
        assert version in {"RGB", "PCA", "MULTISPECTRAL"}
        assert split in {"TRAIN", "EVAL"}

        self.data_folder = config.DATA_DIR
        self.split = split.capitalize()
        self.version = version
        self.transform = transform
        self.cpi = 5

        self.random_gen = random.Random(seed)

        self.pre_image_dir = os.path.join(
            config.DATA_DIR, "imgs_tiles", version, "500_2000/"
        )
        self.post_image_dir = os.path.join(
            config.DATA_DIR, "imgs_tiles", version, "500_2010/"
        )

        self.imgs = []
        self.captions = []
        self.all_captions = []
        self.labels = []

        self.filenames = []

        if self.split == "Train":
            self._load_data("Train_Dubai_CC.json")
        elif self.split == "Eval":
            self._load_data("Test_Dubai_CC.json")
            self._load_data("Validation_Dubai_CC.json")
        else:
            raise ValueError("Invalid split")

        self.imgs = np.array(
            self.imgs
        )  # (n, 2, channels, 50, 50) or if split == "Train" (n * 5, 2, channels, 50, 50)
        self.captions = np.array(self.captions)  # (n * 5)
        self.all_captions = np.array(self.all_captions)  # (n, 5)
        self.labels = np.array(self.labels)  # (n, 1)

        if self.split != "Train":
            assert len(self.imgs) == len(self.all_captions) == len(self.labels)
            assert len(self.all_captions) * self.cpi
        else:
            assert len(self.imgs) == len(self.captions)

        self.dataset_size = int(len(self.imgs))

    def set_offset(self, offset):
        self.offset = offset

    def __getitem__(self, index):
        img = torch.FloatTensor(self.imgs[index] / 255.0)  # (2, channels, 50, 50)

        if self.transform is not None:
            img[0] = self.transform(img[0])
            img[1] = self.transform(img[1])

        if self.split == "Train":
            caption = self.captions[index]  # (1)
            return img, caption
        else:
            offset = self.random_gen.randint(0, self.cpi - 1)
            caption = self.all_captions[index][offset]  # (1)
            label = self.labels[index]  # (1)
            all_captions = self.all_captions[index].tolist()  # (5)
            filename = self.filenames[index]  # (5)

            return img, caption, label, all_captions, filename

    def _load_data(self, filename: str):
        with open(
            os.path.join(
                config.DATA_DIR,
                "captions",
                filename,
            ),
            "r",
        ) as j:
            captions_raw = json.load(j)

            for item in captions_raw["images"]:
                filename = item["filename"]
                sentences = [
                    caption["raw"].strip(" ").rstrip(".")
                    for caption in item["sentences"]
                ]

                if self.version == "MULTISPECTRAL":
                    pre_image = self._load_multispectral_image(
                        self.pre_image_dir + filename
                    )
                    post_image = self._load_multispectral_image(
                        self.post_image_dir + filename
                    )
                else:
                    pre_image = np.array(Image.open(self.pre_image_dir + filename))
                    post_image = np.array(Image.open(self.post_image_dir + filename))

                    pre_image = np.transpose(pre_image, (2, 0, 1))  # whc -> cwh
                    post_image = np.transpose(post_image, (2, 0, 1))  # whc -> cwh

                self.imgs.append(np.stack((pre_image, post_image)))

                self.captions.extend(sentences)
                self.all_captions.append(sentences)

                self.filenames.append(filename)

                if self.split == "Train":
                    for _ in range(self.cpi - 1):
                        self.imgs.append(np.stack((pre_image, post_image)))
                elif self.split == "Eval":
                    self.labels.append(self._get_label(sentences))
                else:
                    raise ValueError("Invalid split")

    def __len__(self):
        return self.dataset_size

    def _get_label(self, captions):
        if self._contains_no_change(captions):
            return 0  # No change
        else:
            return 1  # Change

    def _contains_no_change(self, captions):
        for caption in captions:
            if any(s in caption for s in no_change_sentences):
                return True
        return False

    def _load_multispectral_image(self, image_path):
        with rasterio.open(image_path) as src:
            image = src.read()  # (bands, height, width), no need for transposition
        return image
