import torch
from args import config
from torchvision import transforms
import h5py
from torch.utils.data import Dataset
import json
import os
import numpy as np
import random


levir_cc_train_transform = transforms.Compose(
    [
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ]
)

levir_cc_test_transform = transforms.Compose(
    [
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ]
)


class LevirDataset(Dataset):
    def __init__(self, split, transform=None, seed=42):
        """
        :param split: split, either 'TRAIN' or 'EVAL'
        :param transform: image transform pipeline
        """

        self.data_folder = config.DATA_DIR
        self.data_name = config.DATA_NAME
        self.split = split
        assert self.split in {"TRAIN", "EVAL"}

        self.random_gen = random.Random(seed)
        self.transform = transform

        self._load_data()

        if self.split == "TRAIN":
            assert len(self.imgs) == len(self.captions)
        elif self.split == "EVAL":
            assert len(self.imgs) == len(self.captions) / self.cpi == len(self.labels)

        self.dataset_size = int(len(self.imgs))

    def _load_data(self):
        if self.split == "TRAIN":
            return self._load_train_data()
        elif self.split == "EVAL":
            return self._load_eval_data()
        else:
            raise ValueError(f"Invalid split: {self.split}")

    def _load_train_data(self):
        h5_file = os.path.join(
            self.data_folder, f"{self.split}_IMAGES_{self.data_name}.hdf5"
        )
        with h5py.File(h5_file, "r") as h:
            self.imgs = np.repeat(h["images"], h.attrs["captions_per_image"], axis=0)
            self.cpi = h.attrs["captions_per_image"]

        captions_file = os.path.join(
            self.data_folder, f"{self.split}_CAPTIONS_{self.data_name}.json"
        )
        with open(captions_file, "r") as j:
            self.captions = [caption.strip(" ").rstrip(".") for caption in json.load(j)]

        self.labels = None
        self.filenames = None

    def _load_eval_data(self):
        h5_files = [
            os.path.join(self.data_folder, f"{split}_IMAGES_{self.data_name}.hdf5")
            for split in ["VAL", "TEST"]
        ]
        imgs = np.concatenate([h5py.File(f, "r")["images"] for f in h5_files], axis=0)
        self.cpi = 5

        captions_files = [
            os.path.join(self.data_folder, f"{split}_CAPTIONS_{self.data_name}.json")
            for split in ["VAL", "TEST"]
        ]

        labels_files = [
            os.path.join(self.data_folder, f"{split}_LABELS_{self.data_name}.json")
            for split in ["VAL", "TEST"]
        ]

        filenames_files = [
            os.path.join(self.data_folder, f"{split}_FILENAMES_{self.data_name}.json")
            for split in ["VAL", "TEST"]
        ]

        captions = []
        for f in captions_files:
            captions.extend(
                [caption.strip(" ").rstrip(".") for caption in json.load(open(f, "r"))]
            )

        labels = []
        for f in labels_files:
            labels.extend(json.load(open(f, "r")))

        filenames = []
        for f in filenames_files:
            filenames.extend(json.load(open(f, "r")))

        self.imgs = imgs
        self.captions = captions
        self.labels = labels
        self.filenames = filenames

    def __getitem__(self, index):
        img = torch.FloatTensor(self.imgs[index] / 255.0)

        if self.transform is not None:
            if img.shape == torch.Size([3, 256, 256]):
                img = self.transform(img)
            elif img.shape == torch.Size([2, 3, 256, 256]):
                img[0] = self.transform(img[0])
                img[1] = self.transform(img[1])

        if self.split == "TRAIN":
            caption = self.captions[index]
            return img, caption
        else:
            base_index = int(index * self.cpi)
            offset = self.random_gen.randint(0, self.cpi - 1)
            caption = self.captions[base_index + offset]
            all_captions = self.captions[
                (index * self.cpi) : ((index * self.cpi) + self.cpi)
            ]
            label = self.labels[index]
            filename = self.filenames[index]

            return img, caption, label, all_captions, filename

    def __len__(self):
        return self.dataset_size
