import numpy as np
from evaluation.text_metrics import calculate_text_metrics
from scores import RunningScores, Scores, TextMetrics


class Evaluator:
    def __init__(
        self,
        img_embeddings,
        txt_embeddings,
        captions,
        labels,
        all_captions,
        filenames,
        distance_metric,
        logger,
        k,
    ):
        self.logger = logger
        self.k = k
        self.distance_metric = distance_metric

        self.img_embeddings = img_embeddings
        self.txt_embeddings = txt_embeddings
        self.captions = captions
        self.labels = labels
        self.all_captions = all_captions
        self.filenames = filenames

    def get_metrics(
        self,
        query_img_embeddings,
        query_txt_embeddings,
        query_captions,
        query_all_captions,
        retrieval_img_embeddings,
        retrieval_txt_embeddings,
        retrieval_captions,
        retrieval_all_captions,
    ):
        MTM_I2T, top_k_indices_i2t = calculate_text_metrics(
            query_embeddings=query_img_embeddings,
            retrieval_embeddings=retrieval_txt_embeddings,
            query_captions=query_captions,
            query_all_captions=query_all_captions,
            retrieval_captions=retrieval_captions,
            retrieval_all_captions=retrieval_all_captions,
            top_k=self.k,
            distance_metric=self.distance_metric,
            is_direction_ti=False,
            return_top_k_indices=True,
        )
        MTM_T2I, top_k_indices_t2i = calculate_text_metrics(
            query_embeddings=query_txt_embeddings,
            retrieval_embeddings=retrieval_img_embeddings,
            query_captions=query_captions,
            query_all_captions=query_all_captions,
            retrieval_captions=retrieval_captions,
            retrieval_all_captions=retrieval_all_captions,
            top_k=self.k,
            distance_metric=self.distance_metric,
            is_direction_ti=True,
            return_top_k_indices=True,
        )

        return (
            TextMetrics(**MTM_I2T),
            top_k_indices_i2t,
            TextMetrics(**MTM_T2I),
            top_k_indices_t2i,
        )

    def eval(
        self,
        running_scores: RunningScores,
    ) -> None:
        avg_mtm = Scores()
        avg_mtm_change = Scores()
        avg_mtm_no_change = Scores()

        change_counter = 0

        for i in range(len(self.img_embeddings)):
            query_img_embedding = self.img_embeddings[i].detach().numpy()
            query_txt_embedding = self.txt_embeddings[i].detach().numpy()
            query_label = self.labels[i]  # 1: changed, 0: no change
            query_caption = self.captions[i]
            query_all_captions = self.all_captions[i]
            query_filename = self.filenames[i]
            self.logger.info(
                f"--------------------Query {i + 1}/{len(self.img_embeddings)}, ({'change' if query_label else 'no change'}), {query_filename}-------------------"
            )

            if query_label:
                change_counter += 1

            retrieval_img_embedding = np.delete(
                self.img_embeddings.detach().numpy(), i, axis=0
            )
            retrieval_txt_embedding = np.delete(
                self.txt_embeddings.detach().numpy(), i, axis=0
            )
            retrieval_captions = np.delete(self.captions, i, axis=0)
            retrieval_all_captions = np.delete(self.all_captions, i, axis=0)
            retrieval_filenames = np.delete(self.filenames, i, axis=0)

            (
                MTM_I2T,
                top_k_indices_i2t,
                MTM_T2I,
                top_k_indices_t2i,
            ) = self.get_metrics(
                [query_img_embedding],
                [query_txt_embedding],
                [query_caption],
                [query_all_captions],
                retrieval_img_embeddings=retrieval_img_embedding,
                retrieval_txt_embeddings=retrieval_txt_embedding,
                retrieval_captions=retrieval_captions,
                retrieval_all_captions=retrieval_all_captions,
            )

            avg_mtm += Scores(MTM_I2T, MTM_T2I)

            if query_label:
                avg_mtm_change += Scores(MTM_I2T, MTM_T2I)
            else:
                avg_mtm_no_change += Scores(MTM_I2T, MTM_T2I)

            self.logger.info(
                f"Bleu_1@{self.k} I->T: %.3f, Bleu_1@{self.k} T->I: %.3f"
                % (MTM_I2T.bleu_1, MTM_T2I.bleu_1)
            )

            self.logger.info(
                f"Bleu_2@{self.k} I->T: %.3f, Bleu_2@{self.k} T->I: %.3f"
                % (MTM_I2T.bleu_2, MTM_T2I.bleu_2)
            )

            self.logger.info(
                f"Bleu_3@{self.k} I->T: %.3f, Bleu_3@{self.k} T->I: %.3f"
                % (MTM_I2T.bleu_3, MTM_T2I.bleu_3)
            )

            self.logger.info(
                f"Bleu_4@{self.k} I->T: %.3f, Bleu_4@{self.k} T->I: %.3f"
                % (MTM_I2T.bleu_4, MTM_T2I.bleu_4)
            )

            self.logger.info(
                f"METEOR@{self.k} I->T: %.3f, METEOR@{self.k} T->I: %.3f"
                % (MTM_I2T.meteor, MTM_T2I.meteor)
            )

            self.logger.info(
                f"ROUGE_L@{self.k} I->T: %.3f, ROUGE_L@{self.k} T->I: %.3f"
                % (MTM_I2T.rouge_l, MTM_T2I.rouge_l)
            )

            self.logger.info(
                f"CIDEr@{self.k} I->T: %.3f, CIDEr@{self.k} T->I: %.3f"
                % (MTM_I2T.cider, MTM_T2I.cider)
            )

            self.logger.info(f"Top {self.k} retrieval results (I->T):")
            self.logger.info(
                f"Query: {query_filename}, ({', '.join(query_all_captions)})"
            )
            for i, idx in enumerate(top_k_indices_i2t):
                self.logger.info(
                    f"Top {i + 1}: Caption: {retrieval_captions[idx]} ({retrieval_filenames[idx]})"
                )
            self.logger.info(f"Top {self.k} retrieval results (T->I):")
            self.logger.info(f"Query: {query_caption} ({query_filename})")
            for i, idx in enumerate(top_k_indices_t2i):
                self.logger.info(
                    f"Top {i + 1}: Filename: {retrieval_filenames[idx]} ({', '.join(retrieval_all_captions[idx])}) "
                )

        self.logger.info(
            "--------------------Average Scores in Evaluation Round-------------------"
        )

        avg_mtm /= len(self.img_embeddings)
        avg_mtm_change /= change_counter
        avg_mtm_no_change /= len(self.img_embeddings) - change_counter

        running_scores.full_set.append(avg_mtm)
        running_scores.change_set.append(avg_mtm_change)
        running_scores.nochange_set.append(avg_mtm_no_change)

        self.logger.info("Full set scores")
        self.logger.info(avg_mtm)

        self.logger.info("Change set scores")
        self.logger.info(avg_mtm_change)

        self.logger.info("No change set scores")
        self.logger.info(avg_mtm_no_change)
