import numpy as np

from args import config

from nlgeval import compute_metrics

from utils import calculate_hamming_distance, calculate_cosine_similarity


def calculate_text_metrics(
    query_embeddings,
    retrieval_embeddings,
    query_captions,
    query_all_captions,
    retrieval_captions,
    retrieval_all_captions,
    top_k,
    is_direction_ti,
    distance_metric="hamming",
    return_top_k_indices=False,
):
    """
    query_embeddings: query embeddings, (query set size x embedding dim)
    retrieval_embeddings: retrieval embeddings, (archive size x embedding dim)
    query_captions: query captions, (query set size)
    query_all_captions: query captions, (query set size, caption per image)
    retrieval_captions: retrieval captions, (archive size)
    retrieval_all_captions: retrieval captions, (archive size, caption per image)
    """

    assert len(query_captions) == len(query_all_captions)
    assert len(retrieval_captions) == len(retrieval_all_captions)
    assert len(retrieval_all_captions[0]) == len(query_all_captions[0]) == 5

    num_queries = len(query_embeddings)

    query_embeddings = np.stack(query_embeddings)

    hypotheses = []
    references_dict = {
        0: [],
        1: [],
        2: [],
        3: [],
        4: [],
    }

    # I -> T: for the query we need all 5 captions (so 50*5) for reference, and for the hypothesis we need only the captions that have been used for the code generation in the retrieval (so 150*1)
    # T -> I: for the query we need 1 caption (so 50*1) for the hypothesis, and for the reference, we need all 5 captions of the images (so 150*5)

    # query codes in T->I direction is a list of 5 * batch_size codes and they are compared to the codes of the 150 images, so that is fine
    # query codes in the I->T is a list of batch_size codes and they are compared to the codes of the 150 * 5 text
    for i in range(num_queries):
        # compute distance between query and retrieval codes
        query_embedding = query_embeddings[i, :]
        if distance_metric == "cosine":
            distances = calculate_cosine_similarity(
                query_embedding, retrieval_embeddings
            )
            idx = np.argsort(distances)[::-1]  # sort by similarity
        elif distance_metric == "hamming":
            distances = calculate_hamming_distance(
                query_embedding, retrieval_embeddings
            )
            idx = np.argsort(distances)  # sort by distance

        if is_direction_ti:
            # hypothesis is the text query caption (even though it is not generated), reference captions are the captions of retrieved images
            # we will compare the different (retrieved) references with the same (query) hypothesis

            references = retrieval_all_captions[idx][
                0:top_k
            ]  # top_kx5 retrieved captions -> contains the 5 captions for each of the top_k retrieved images

            hypothesis = query_captions[
                i
            ]  # contains the actual query caption that has been used to generate the hash code

            for _ in range(top_k):
                hypotheses.append(
                    hypothesis
                )  # we add the same hypothesis top_k times because we compare it with top_k different references (top_k retrieved batches of 5 sentences)

            for (
                ref_batch
            ) in references:  # we add each reference batch (top_k * 5 sentences)
                for j, ref in enumerate(ref_batch):
                    references_dict[j].append(ref)

        else:
            # hypotheses are the retrieved text captions that have been "generated" by the system for a given query image
            # we use the query image captions as references and compare the different (retrieved) hypotheses with the five (query) reference

            references = query_all_captions[
                i
            ]  # contains the 5 captions of the query image
            hypos = retrieval_captions[idx][
                0:top_k
            ]  # contains the caption of the top_k retrieved items that have been used

            hypotheses.extend(hypos)

            for _ in range(top_k):
                for j, ref in enumerate(references):
                    references_dict[j].append(ref)

    return (
        get_text_metrics(list(references_dict.values()), hypotheses)
        if not return_top_k_indices
        else (
            get_text_metrics(list(references_dict.values()), hypotheses),
            idx[0:top_k],
        )
    )


def get_text_metrics(references, hypotheses):
    text_metrics_dict = compute_metrics(
        hypothesis=hypotheses,
        references=references,
        no_overlap=False,
        no_skipthoughts=True,
        no_glove=True,
    )

    return text_metrics_dict
