import torch.nn as nn
import torch
from multispectral_adaptation import modify_model_for_multispectral
import clip.clip as clip
import torch.nn.functional as F
from cross_attention import MCCFormers_diff_as_Q

import torchvision.models as models
from torchvision.models import ResNet50_Weights

clip_model, _ = clip.load(
    "ViT-B/16", device="cuda:0" if torch.cuda.is_available() else "cpu"
    # ViT-B/32
)

device = "cuda" if torch.cuda.is_available() else "cpu"

clip_token_dim = 768


class ImgNet(nn.Module):
    def __init__(
        self, input_size, use_resnet=False, operation="subtract", version="RGB"
    ):
        super(ImgNet, self).__init__()
        self.use_resnet = use_resnet

        assert version in ["RGB", "PCA", "MULTISPECTRAL"]
        self.version = version

        assert operation in ["subtract", "concat", "add"]
        self.operation = operation

        if self.use_resnet:
            self.resnet = models.resnet50(weights=ResNet50_Weights.DEFAULT)
            # self.resnet.eval() #try

            if self.version == "MULTISPECTRAL":
                self.modify_resnet_for_multispectral()

            self.resnet.fc = nn.Identity()
        else:
            self.clip_model = clip_model

            if self.version == "MULTISPECTRAL":
                self.modify_clip_for_multispectral()

        if self.operation == "concat":
            self.fc1 = nn.Linear(input_size * 2, 256)
        else:
            self.fc1 = nn.Linear(input_size, 256)

        self.fc2 = nn.Linear(256, 128)

    def modify_resnet_for_multispectral(self):
        self.resnet.conv1 = modify_model_for_multispectral(self.resnet.conv1, 6)

    def modify_clip_for_multispectral(self):
        self.clip_model.visual.conv1 = modify_model_for_multispectral(
            self.clip_model.visual.conv1, 6
        )

    def forward(self, feat_I):
        x_0 = feat_I[:, 0, :, :, :]
        x_1 = feat_I[:, 1, :, :, :]

        with torch.no_grad():
            x_0 = F.interpolate(
                x_0, size=(224, 224), mode="bilinear", align_corners=False
            )
            x_1 = F.interpolate(
                x_1, size=(224, 224), mode="bilinear", align_corners=False
            )

            if self.use_resnet:
                x_0_enc = self.resnet(x_0)
                x_1_enc = self.resnet(x_1)
            else:
                x_0_enc = self.clip_model.encode_image(x_0)
                x_0_enc = x_0_enc.type(torch.float32)
                x_1_enc = self.clip_model.encode_image(x_1)
                x_1_enc = x_1_enc.type(torch.float32)

        if self.operation == "subtract":
            feat_I = x_1_enc - x_0_enc
        elif self.operation == "add":
            feat_I = x_1_enc + x_0_enc
        elif self.operation == "concat":
            feat_I = torch.cat((x_0_enc, x_1_enc), dim=1)

        feat_I = torch.relu(self.fc1(feat_I))
        feat_I = self.fc2(feat_I)

        return feat_I


class TxtNetClip(nn.Module):
    def __init__(self, txt_feat_len):
        super(TxtNetClip, self).__init__()

        self.fc1 = nn.Linear(txt_feat_len, 256)
        self.fc2 = nn.Linear(256, 128)

    def forward(self, x):
        with torch.no_grad():
            x = torch.cat([clip.tokenize(caption) for caption in x])  # [batch_size, 77]
            feat_T = clip_model.encode_text(x.to(device))
            feat_T = feat_T.type(torch.float32)

        feat_T = torch.relu(self.fc1(feat_T))
        feat_T = self.fc2(feat_T)

        return feat_T


# requires last lines in VisionTransformer forward method to be removed from model.py (CLIP)

class ImgNetFusion(nn.Module):
    def __init__(self, input_size, use_resnet=False, version="RGB"):
        super(ImgNetFusion, self).__init__()
        self.use_resnet = use_resnet

        assert version in ["RGB", "PCA", "MULTISPECTRAL"]
        self.version = version

        if self.use_resnet:
            self.resnet = models.resnet50(weights=ResNet50_Weights.DEFAULT)

            if self.version == "MULTISPECTRAL":
                print ('entra??')
                self.modify_resnet_for_multispectral()
            self.resnet = torch.nn.Sequential(*(list(self.resnet.children())[:-3]))

            self.feature_dim = 1024
        else:
            self.clip_model = clip_model
            self.feature_dim = clip_token_dim

            if self.version == "MULTISPECTRAL":
                self.modify_clip_for_multispectral()

        self.encoder_feat = MCCFormers_diff_as_Q(
            feature_dim=self.feature_dim,
            dropout=0.2,
            h=14,
            w=14,
            d_model=512,
            n_head=16,  # args.n_heads,
            n_layers=3,  # args.n_layers,
        )

        self.fc1 = nn.Linear(input_size, 512)
        self.fc2 = nn.Linear(512, 128)


    def modify_resnet_for_multispectral(self):

        self.resnet.conv1 = modify_model_for_multispectral(self.resnet.conv1, 6)

    def modify_clip_for_multispectral(self):
        self.clip_model.visual.conv1 = modify_model_for_multispectral(
            self.clip_model.visual.conv1, 6
        )

    def forward(self, feat_I):
        x_0 = feat_I[:, 0, :, :, :]
        x_1 = feat_I[:, 1, :, :, :]

        with torch.no_grad():
            x_0 = F.interpolate(
                x_0, size=(224, 224), mode="bilinear", align_corners=False
            )
            x_1 = F.interpolate(
                x_1, size=(224, 224), mode="bilinear", align_corners=False
            )

            if self.use_resnet:
                x_0_enc = self.resnet(x_0)  # (batch_size, 1024, 14, 14)
                x_1_enc = self.resnet(x_1)  # (batch_size, 1024, 14, 14)

                x_0_enc = x_0_enc.view(
                    x_0_enc.size(0), 1024, -1
                )  # (batch_size, 1024, 196)
                x_1_enc = x_1_enc.view(
                    x_1_enc.size(0), 1024, -1
                )  # (batch_size, 1024, 196)

                x_0_enc = x_0_enc.permute(0, 2, 1)  # (batch_size, 196, 1024)
                x_1_enc = x_1_enc.permute(0, 2, 1)  # (batch_size, 196, 1024)
            else:
                x_0_enc = self.clip_model.encode_image(
                    x_0
                )  # (batch, patches, token_repr_dim) (20, 197, 768)
                x_0_enc = x_0_enc.type(torch.float32)
                x_1_enc = self.clip_model.encode_image(
                    x_1
                )  # (batch, patches, token_repr_dim) (20, 197, 768)
                x_1_enc = x_1_enc.type(torch.float32)

                # exclude CLS token
                x_0_enc = x_0_enc[:, 1:, :]
                x_1_enc = x_1_enc[:, 1:, :]
        # print("x_0_enc shape:", x_0_enc.shape)
        # print("x_1_enc shape:", x_1_enc.shape)               
        # import ipdb; ipdb.set_trace()

        feat_I = self.encoder_feat(
            x_0_enc, x_1_enc
        )  # (num_patches, batch, feature_dim) [196, 20, 1024]

        feat_I = feat_I.permute(1, 0, 2)
        feat_I = feat_I.mean(dim=1)

        feat_I = torch.relu(self.fc1(feat_I))  # (batch, 512)
        # feat_I = torch.relu(self.fc2(feat_I))  # (batch, 256)
        feat_I = self.fc2(feat_I)  # (batch, 128)

        return feat_I
