import torch
import numpy as np
import os.path as osp
import torch.nn.functional as F
from torch.utils.data import DataLoader
from datasets.dubai_cc import (
    DubaiDataset,
    dubai_cc_train_transform,
    dubai_cc_test_transform,
)
from datasets.levir_cc import (
    LevirDataset,
    levir_cc_train_transform,
    levir_cc_test_transform,
)
from datasets.levir_cc_small import (
    LevirSmallDataset,
    levir_cc_small_test_transform,
    levir_cc_small_train_transform,
)
from datasets.utils import collate_fn
from evaluation.evaluator import Evaluator
from models import TxtNetClip, ImgNet, ImgNetFusion
from scores import RunningScores, Scores
from plotting.save_figures import save_figures


class MyModel:
    def __init__(self, log, config):
        self.logger = log
        self.config = config
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        torch.manual_seed(1)

        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(1)
            torch.cuda.set_device(self.config.GPU_ID)

        if self.config.DATASET == "LEVIR-CC":
            self.train_dataset = LevirDataset(
                "TRAIN",
                transform=levir_cc_train_transform,
            )
            self.eval_dataset = LevirDataset(
                "EVAL",
                transform=levir_cc_test_transform,
            )

        elif self.config.DATASET == "DUBAI-CC":
            self.train_dataset = DubaiDataset(
                "TRAIN",
                version=config.VERSION,
                transform=dubai_cc_train_transform[config.VERSION],
            )
            self.eval_dataset = DubaiDataset(
                "EVAL",
                version=config.VERSION,
                transform=dubai_cc_test_transform[config.VERSION],
            )

        elif self.config.DATASET == "LEVIR-CC-SMALL":
            self.train_dataset = LevirSmallDataset(
                "TRAIN",
                transform=levir_cc_small_train_transform,
            )
            self.eval_dataset = LevirSmallDataset(
                "EVAL",
                transform=levir_cc_small_test_transform,
            )

        self.train_loader = DataLoader(
            dataset=self.train_dataset,
            batch_size=self.config.BATCH_SIZE,
            shuffle=True,
            num_workers=self.config.NUM_WORKERS,
            drop_last=True,
        )

        self.eval_loader = DataLoader(
            dataset=self.eval_dataset,
            batch_size=self.config.BATCH_SIZE,
            shuffle=False,
            collate_fn=collate_fn,
            num_workers=self.config.NUM_WORKERS,
        )

        if self.config.IMG_NET == "resnet":
            if config.FUSION_STRATEGY == "fusion":
                self.img_net = ImgNetFusion(
                    input_size=1024, use_resnet=True, version=config.VERSION
                )
            else:
                self.img_net = ImgNet(
                    input_size=2048,
                    use_resnet=True,
                    operation=self.config.FUSION_STRATEGY,
                    version=config.VERSION,
                )
        elif self.config.IMG_NET == "clip":
            if config.FUSION_STRATEGY == "fusion":
                self.img_net = ImgNetFusion(
                    input_size=1024, use_resnet=False, version=config.VERSION
                )
            else:
                self.img_net = ImgNet(
                    input_size=512,
                    use_resnet=False,
                    operation=self.config.FUSION_STRATEGY,
                    version=config.VERSION,
                )
        else:
            raise ValueError(
                "Invalid IMG_NET value. Configuration must be 'resnet' or 'clip'"
            )
        self.txt_net = TxtNetClip(txt_feat_len=512)  # resnet, clip, fusion: 512

        self.logit_scale = torch.nn.Parameter(
            torch.ones([]) * np.log(1 / 0.07)
        )  # initialize as in CLIP

        self.opt_img = torch.optim.SGD(
            list(self.img_net.parameters()) + [self.logit_scale],
            lr=self.config.LR_IMG,
            momentum=self.config.MOMENTUM,
            weight_decay=self.config.WEIGHT_DECAY,
        )
        self.opt_txt = torch.optim.SGD(
            list(self.txt_net.parameters()) + [self.logit_scale],
            lr=self.config.LR_TXT,
            momentum=self.config.MOMENTUM,
            weight_decay=self.config.WEIGHT_DECAY,
        )

        self.running_scores = [RunningScores() for _ in range(self.config.EVAL_ROUNDS)]
        self.best_scores = Scores()
        self.std_best_scores = Scores()
        self.losses = []

    def train(self, epoch):
        self.img_net.to(self.device).train()
        self.txt_net.to(self.device).train()

        for idx, (img_pairs, txt) in enumerate(self.train_loader):
            img_pairs = img_pairs.to(self.device)

            self.opt_img.zero_grad()
            self.opt_txt.zero_grad()

            F_I = self.img_net(img_pairs)
            F_T = self.txt_net(txt)

            loss = self._contrastive_loss(F_I, F_T)

            loss.backward()

            self.opt_img.step()
            self.opt_txt.step()

            if (idx + 1) % (
                len(self.train_dataset)
                // self.config.BATCH_SIZE
                / self.config.EPOCH_INTERVAL
            ) == 0:
                self.losses.append(loss.item())
                self.logger.info(
                    "Epoch [%d/%d], Iter [%d/%d] Loss: %.4f"
                    % (
                        epoch + 1,
                        self.config.NUM_EPOCH,
                        idx + 1,
                        len(self.train_dataset) // self.config.BATCH_SIZE,
                        loss.item(),
                    )
                )

    def compute_embeddings(self):
        self.img_net.to(self.device).eval()
        self.txt_net.to(self.device).eval()

        img_embeddings: list = []
        txt_embeddings: list = []
        imgs: list = []
        label_lst: list = []
        captions_lst: list = []
        all_captions_lst: list = []
        filenames_lst: list = []

        for img_pairs, captions, labels, all_captions, filenames in self.eval_loader:
            img_embedding = self.img_net(img_pairs.to(self.device))
            txt_embedding = self.txt_net(captions)

            img_embeddings.append(img_embedding.detach().cpu())
            txt_embeddings.append(txt_embedding.detach().cpu())

            imgs.append(img_pairs.cpu())
            label_lst.append(labels.cpu())
            captions_lst.extend(captions)
            all_captions_lst.extend(all_captions)
            filenames_lst.extend(filenames)

        img_embeddings = torch.cat(img_embeddings, dim=0)
        txt_embeddings = torch.cat(txt_embeddings, dim=0)
        imgs = torch.cat(imgs, dim=0)
        label_lst = torch.cat(label_lst, dim=0)

        captions_lst = np.array(captions_lst)
        all_captions_lst = np.array(all_captions_lst)

        return (
            img_embeddings,
            txt_embeddings,
            imgs,
            captions_lst,
            label_lst,
            all_captions_lst,
            filenames_lst,
        )

    def eval(self):
        for i in range(self.config.EVAL_ROUNDS):
            self.logger.info(f"Evaluation Round {i + 1}, k={self.config.topk}")
            (
                img_embeddings,
                txt_embeddings,
                _,
                captions,
                labels,
                all_captions,
                filenames,
            ) = self.compute_embeddings()

            evaluator = Evaluator(
                img_embeddings,
                txt_embeddings,
                captions,
                labels,
                all_captions,
                filenames,
                distance_metric="cosine",
                logger=self.logger,
                k=self.config.topk,
            )

            evaluator.eval(self.running_scores[i])

        mean_scores = sum(self.running_scores) / len(self.running_scores)
        std_scores = (
            sum([(x - mean_scores) ** 2 for x in self.running_scores])
            / len(self.running_scores)
        ) ** 0.5

        self.logger.info(
            "============== Mean of evaluation rounds in this epoch =============="
        )
        self.logger.info(str(mean_scores))
        self.logger.info(
            "============== Std of evaluation rounds in this epoch =============="
        )
        self.logger.info(str(std_scores))

        if mean_scores.full_set[-1] > self.best_scores:
            self.best_scores = mean_scores.full_set[-1]
            self.std_best_scores = std_scores.full_set[-1]
            self.logger.info("New best scores!")
            self.logger.info(str(self.best_scores))

        self.logger.info(
            "Mean Scores From Running Scores (results of each epoch in order)"
        )
        self.logger.info(str(mean_scores))

        return mean_scores.full_set[-1]

    def print_results(self):
        self.logger.info(
            "--------------------------------------------------------------------"
        )

        mean_scores = sum(self.running_scores) / len(self.running_scores)
        std_scores = (
            sum([(x - mean_scores) ** 2 for x in self.running_scores])
            / len(self.running_scores)
        ) ** 0.5

        self.logger.info("Latex Full Set Scores")
        self.logger.info(mean_scores.full_set[-1].get_latex())
        self.logger.info("Latex Std Full Set Scores")
        self.logger.info(std_scores.full_set[-1].get_latex())

        self.logger.info("Latex Change Set Scores")
        self.logger.info(mean_scores.change_set[-1].get_latex())
        self.logger.info("Latex Std Change Set Scores")
        self.logger.info(std_scores.change_set[-1].get_latex())

        self.logger.info("Latex No Change Set Scores")
        self.logger.info(mean_scores.nochange_set[-1].get_latex())
        self.logger.info("Latex Std No Change Set Scores")
        self.logger.info(std_scores.nochange_set[-1].get_latex())

        self.logger.info(
            "--------------------------------------------------------------------"
        )

    def save_figures(self):
        save_figures(self.losses)

    def _contrastive_loss(self, image_features, text_features):
        image_features = image_features / image_features.norm(dim=1, keepdim=True)
        text_features = text_features / text_features.norm(dim=1, keepdim=True)

        logit_scale = self.logit_scale.exp()
        logits_per_image = (
            logit_scale * image_features @ text_features.t()
        )  # [batch_size, batch_size]
        logits_per_text = logits_per_image.t()

        # labels are indices of diagonal elements since img and txt are aligned
        ground_truth = torch.arange(logits_per_image.shape[0], device=self.device)

        return (
            F.cross_entropy(logits_per_image, ground_truth)
            + F.cross_entropy(logits_per_text, ground_truth) / 2
        )

    def save_checkpoints(self, file_name="latest.pth"):
        ckp_path = osp.join(self.config.CHECKPOINT_DIR, file_name)
        obj = {
            "ImgNet": self.img_net.state_dict(),
            "TxtNet": self.txt_net.state_dict(),
        }
        torch.save(obj, ckp_path)
        self.logger.info("**********Saved the model successfully.**********")

    def load_checkpoints(self, file_name="latest.pth"):
        ckp_path = osp.join(self.config.CHECKPOINT_DIR, file_name)
        try:
            obj = torch.load(
                ckp_path,
                map_location=lambda storage, loc: (
                    storage.cuda() if torch.cuda.is_available() else storage
                ),
            )
            self.img_net.load_state_dict(obj["ImgNet"])
            self.txt_net.load_state_dict(obj["TxtNet"])
            self.logger.info(
                "**************** Load checkpoint %s ****************" % ckp_path
            )
        except IOError:
            self.logger.error(
                "********** Fail to load checkpoint %s!*********" % ckp_path
            )
            raise IOError
