import numpy as np
import logging
import os.path as osp
import os

from args import config
import time


def calculate_hamming_distance(query_binary_vector, retrieval_binary_matrix):
    """
    Computes the Hamming distance between a binary query vector and each binary vector in the retrieval matrix.

    :param query_code: binary vector of the query [n].
    :param retrieval_codes: binary matrix of the retrieval items [r*n].
    :return: array of Hamming distances for each retrieval item [r].
    """

    # maximum inner product value is the length of the vectors
    vector_length = retrieval_binary_matrix.shape[1]
    inner_products = np.dot(query_binary_vector, retrieval_binary_matrix.transpose())

    # Hamming distance is defined as half the difference between the maximum
    # possible inner product (vector length) and the actual inner product.
    hamming_distances = 0.5 * (vector_length - inner_products)

    return hamming_distances


def calculate_cosine_similarity(query_embedding, retrieval_embeddings):
    """
    Computes the cosine similarity between a query vector and each vector in the retrieval matrix.

    :param query_embedding: vector of the query [n].
    :param retrieval_embeddings: matrix of the retrieval items [r*n].
    :return: array of cosine similarities for each retrieval item [r].
    """
    # Normalize the vectors
    query_embedding_norm = query_embedding / np.linalg.norm(query_embedding)
    retrieval_embeddings_norm = retrieval_embeddings / np.linalg.norm(
        retrieval_embeddings, axis=1, keepdims=True
    )

    # Compute the cosine similarity
    cosine_similarities = np.dot(query_embedding_norm, retrieval_embeddings_norm.T)

    return cosine_similarities


def create_run_dir():
    if config.DATASET == "LEVIR-CC":
        dataset_name = "levir_cc"
    elif config.DATASET == "LEVIR-CC-SMALL":
        dataset_name = "levir_cc_small"
    elif config.DATASET == "DUBAI-CC":
        dataset_name = "dubai_" + config.VERSION.lower()
    else:
        raise ValueError("Invalid dataset")

    root = f"./runs/{dataset_name}_{config.MODEL_NAME}_{time.strftime('%Y-%m-%d_%H-%M-%S')}"

    log_dir = osp.join(root, "logs")
    checkpoint_dir = osp.join(root, "checkpoints")
    figure_dir = osp.join(root, "figures")

    os.makedirs(root)
    os.makedirs(checkpoint_dir)
    os.makedirs(log_dir)
    os.makedirs(figure_dir)

    config.LOG_DIR = log_dir
    config.CHECKPOINT_DIR = checkpoint_dir
    config.FIGURE_DIR = figure_dir


def logger():
    logger = logging.getLogger("train")

    logger.setLevel(logging.INFO)

    log_name = "logs.log"
    log_dir = config.LOG_DIR
    txt_log = logging.FileHandler(osp.join(log_dir, log_name))
    txt_log.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    txt_log.setFormatter(formatter)
    logger.addHandler(txt_log)

    stream_log = logging.StreamHandler()
    stream_log.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    stream_log.setFormatter(formatter)
    logger.addHandler(stream_log)

    return logger
